"""Tests for the `PredictPayees` and the `PredictPostings` decorator"""

# pylint: disable=missing-docstring
from beancount.parser import parser
from beangulp import Importer

from smart_importer import PredictPayees, PredictPostings

TEST_DATA, _, __ = parser.parse_string(
    """
2017-01-06 * "Farmer Fresh" "Buying groceries"
  Assets:US:BofA:Checking  -2.50 USD

2017-01-07 * "Groceries"
  Assets:US:BofA:Checking  -10.20 USD

2017-01-10 * "" "Eating out with Joe"
  Assets:US:BofA:Checking  -38.36 USD

2017-01-10 * "Dinner with Martin"
  Assets:US:BofA:Checking  -35.00 USD

2017-01-10 * "Groceries"
  Assets:US:BofA:Checking  -53.70 USD

2017-01-10 * "Gimme Coffee" "Coffee"
  Assets:US:BofA:Checking  -5.00 USD

2017-01-12 * "Uncle Boons" ""
  Assets:US:BofA:Checking  -27.00 USD

2017-01-13 * "Gas Quick"
  Assets:US:BofA:Checking  -17.45 USD

2017-01-14 * "Axe Throwing with Joe"
  Assets:US:BofA:Checking  -13.37 USD
"""
)

TRAINING_DATA, _, __ = parser.parse_string(
    """
2016-01-01 open Assets:US:BofA:Checking USD
2016-01-01 open Expenses:Food:Coffee USD
2016-01-01 open Expenses:Auto:Diesel USD
2016-01-01 open Expenses:Auto:Gas USD
2016-01-01 open Expenses:Food:Groceries USD
2016-01-01 open Expenses:Food:Restaurant USD
2016-01-01 open Expenses:Denylisted USD

2016-01-06 * "Farmer Fresh" "Buying groceries"
  Assets:US:BofA:Checking  -2.50 USD
  Expenses:Food:Groceries

2016-01-07 * "Starbucks" "Coffee"
  Assets:US:BofA:Checking  -4.00 USD
  Expenses:Food:Coffee

2016-01-07 * "Farmer Fresh" "Groceries"
  Assets:US:BofA:Checking  -10.20 USD
  Expenses:Food:Groceries

2016-01-07 * "Gimme Coffee" "Coffee"
  Assets:US:BofA:Checking  -3.50 USD
  Expenses:Food:Coffee

2016-01-07 * "Gas Quick"
  Assets:US:BofA:Checking  -22.79 USD
  Expenses:Auto:Diesel

2016-01-08 * "Uncle Boons" "Eating out with Joe"
  Assets:US:BofA:Checking  -38.36 USD
  Expenses:Food:Restaurant

2016-01-10 * "Walmarts" "Groceries"
  Assets:US:BofA:Checking  -53.70 USD
  Expenses:Food:Groceries

2016-01-10 * "Gimme Coffee" "Coffee"
  Assets:US:BofA:Checking  -6.19 USD
  Expenses:Food:Coffee

2016-01-10 * "Gas Quick"
  Assets:US:BofA:Checking  -21.60 USD
  Expenses:Auto:Diesel

2016-01-10 * "Uncle Boons" "Dinner with Mary"
  Assets:US:BofA:Checking  -35.00 USD
  Expenses:Food:Restaurant

2016-01-11 close Expenses:Auto:Diesel

2016-01-11 * "Farmer Fresh" "Groceries"
  Assets:US:BofA:Checking  -30.50 USD
  Expenses:Food:Groceries

2016-01-12 * "Gas Quick"
  Assets:US:BofA:Checking  -24.09 USD
  Expenses:Auto:Gas

2016-01-08 * "Axe Throwing with Joe"
  Assets:US:BofA:Checking  -38.36 USD
  Expenses:Denylisted

"""
)

PAYEE_PREDICTIONS = [
    "Farmer Fresh",
    "Farmer Fresh",
    "Uncle Boons",
    "Uncle Boons",
    "Farmer Fresh",
    "Gimme Coffee",
    "Uncle Boons",
    None,
    None,
]

ACCOUNT_PREDICTIONS = [
    "Expenses:Food:Groceries",
    "Expenses:Food:Groceries",
    "Expenses:Food:Restaurant",
    "Expenses:Food:Restaurant",
    "Expenses:Food:Groceries",
    "Expenses:Food:Coffee",
    "Expenses:Food:Groceries",
    "Expenses:Auto:Gas",
    "Expenses:Food:Groceries",
]

DENYLISTED_ACCOUNTS = ["Expenses:Denylisted"]


class DummyImporter(Importer):
    def identify(self, filepath):
        return True

    def account(self, filepath):
        return "Assets:US:BofA:Checking"

    def extract(self, filepath, existing):
        return TEST_DATA


def create_dummy_imports(data):
    return [("file", data, "Assets:US:BofA:Checking", "importer")]


def test_empty_training_data():
    """
    Verifies that the decorator leaves the narration intact.
    """
    assert (
        PredictPayees().hook(create_dummy_imports(TEST_DATA), [])[0][1]
        == TEST_DATA
    )
    assert (
        PredictPostings().hook(create_dummy_imports(TEST_DATA), [])[0][1]
        == TEST_DATA
    )


def test_no_transactions() -> None:
    """
    Should not crash when passed empty list of transactions.
    """
    PredictPayees().hook([], [])
    PredictPostings().hook([], [])
    PredictPayees().hook([], TRAINING_DATA)
    PredictPostings().hook([], TRAINING_DATA)
    PredictPayees().hook(create_dummy_imports([]), TRAINING_DATA)
    PredictPostings().hook(create_dummy_imports([]), TRAINING_DATA)


def test_unchanged_narrations() -> None:
    """
    Verifies that the decorator leaves the narration intact
    """
    correct_narrations = [transaction.narration for transaction in TEST_DATA]
    extracted_narrations = [
        transaction.narration
        for transaction in PredictPayees().hook(
            create_dummy_imports(TEST_DATA), TRAINING_DATA
        )[0][1]
    ]
    assert extracted_narrations == correct_narrations


def test_unchanged_first_posting() -> None:
    """
    Verifies that the decorator leaves the first posting intact
    """
    correct_first_postings = [
        transaction.postings[0] for transaction in TEST_DATA
    ]
    extracted_first_postings = [
        transaction.postings[0]
        for transaction in PredictPayees().hook(
            create_dummy_imports(TEST_DATA), TRAINING_DATA
        )[0][1]
    ]
    assert extracted_first_postings == correct_first_postings


def test_payee_predictions() -> None:
    """
    Verifies that the decorator adds predicted postings.
    """
    transactions = PredictPayees().hook(
        create_dummy_imports(TEST_DATA), TRAINING_DATA
    )[0][1]
    predicted_payees = [transaction.payee for transaction in transactions]
    assert predicted_payees == PAYEE_PREDICTIONS


def test_account_predictions() -> None:
    """
    Verifies that the decorator adds predicted postings.
    """
    predicted_accounts = [
        entry.postings[-1].account
        for entry in PredictPostings(
            denylist_accounts=DENYLISTED_ACCOUNTS
        ).hook(create_dummy_imports(TEST_DATA), TRAINING_DATA)[0][1]
    ]
    assert predicted_accounts == ACCOUNT_PREDICTIONS


def test_account_predictions_wrap() -> None:
    """
    Verifies account prediction using the wrap method instead of the beangulp hook
    """
    wrapped_importer = PredictPostings(
        denylist_accounts=DENYLISTED_ACCOUNTS
    ).wrap(DummyImporter())
    entries = wrapped_importer.extract("dummyFile", TRAINING_DATA)
    print(entries)
    predicted_accounts = [entry.postings[-1].account for entry in entries]
    assert predicted_accounts == ACCOUNT_PREDICTIONS


def test_account_predictions_multiple() -> None:
    """
    Verifies that it's possible to predict multiple importer results
    """
    import_results = [
        ("file1", TEST_DATA, "Assets:US:BofA:Checking", "importer1"),
        ("file1", TEST_DATA, "Assets:US:BofA:Checking", "importer2"),
    ]
    predicted_results = PredictPostings(
        denylist_accounts=DENYLISTED_ACCOUNTS
    ).hook(import_results, TRAINING_DATA)

    assert len(predicted_results) == 2
    predicted_accounts1 = [
        entry.postings[-1].account for entry in predicted_results[0][1]
    ]
    predicted_accounts2 = [
        entry.postings[-1].account for entry in predicted_results[1][1]
    ]
    assert predicted_accounts1 == ACCOUNT_PREDICTIONS
    assert predicted_accounts2 == ACCOUNT_PREDICTIONS



"""Tests for the `PredictPostings` decorator"""

from __future__ import annotations

# pylint: disable=missing-docstring
import os
import pprint
import re
from typing import Callable

import pytest
from beancount.core import data
from beancount.core.compare import stable_hash_namedtuple
from beancount.parser import parser

from smart_importer import PredictPostings


def chinese_string_tokenizer(pre_tokenizer_string: str) -> list[str]:
    jieba = pytest.importorskip("jieba")
    jieba.initialize()
    return list(jieba.cut(pre_tokenizer_string))


def _hash(entry: data.Directive) -> str:
    return stable_hash_namedtuple(entry, ignore={"meta", "units"})


def _load_testset(
    testset: str,
) -> tuple[data.Directives, data.Directives, data.Directives]:
    path = os.path.join(
        os.path.dirname(__file__), "data", testset + ".beancount"
    )
    with open(path, encoding="utf-8") as test_file:
        _, *sections = re.split(r"# [A-Z]+\n", test_file.read())
    parsed_sections = []
    for section in sections:
        entries, errors, __ = parser.parse_string(section)
        assert not errors
        parsed_sections.append(entries)
    assert len(parsed_sections) == 3
    return tuple(parsed_sections)


@pytest.mark.parametrize(
    "testset, account, string_tokenizer",
    [
        ("simple", "Assets:US:BofA:Checking", None),
        ("single-account", "Assets:US:BofA:Checking", None),
        ("multiaccounts", "Assets:US:EUR", None),
        ("chinese", "Assets:US:BofA:Checking", chinese_string_tokenizer),
    ],
)
def test_testset(
    testset: str, account: str, string_tokenizer: Callable[[str], list[str]]
) -> None:
    # pylint: disable=unbalanced-tuple-unpacking
    imported, training_data, expected = _load_testset(testset)

    imported_transactions = PredictPostings(
        string_tokenizer=string_tokenizer
    ).hook([("file", imported, account, "importer")], training_data)

    for txn1, txn2 in zip(imported_transactions[0][1], expected):
        if _hash(txn1) != _hash(txn2):
            pprint.pprint(txn1)
            pprint.pprint(txn2)
            assert False



"""Helpers to work with Beancount entry objects."""

from __future__ import annotations

from beancount.core.data import Posting, Transaction


def update_postings(
    transaction: Transaction, accounts: list[str]
) -> Transaction:
    """Update the list of postings of a transaction to match the accounts.

    Expects the transaction to be updated to have exactly one posting,
    otherwise it is returned unchanged. Adds empty postings for all the
    accounts - if the account of the single existing posting is found
    in the list of accounts, it is placed there at the first occurence,
    otherwise it is appended at the end.
    """

    if len(transaction.postings) != 1:
        return transaction

    posting = transaction.postings[0]

    new_postings = [
        Posting(account, None, None, None, None, None) for account in accounts
    ]
    if posting.account in accounts:
        new_postings[accounts.index(posting.account)] = posting
    else:
        new_postings.append(posting)

    return transaction._replace(postings=new_postings)


def set_entry_attribute(entry, attribute, value, overwrite=False):
    """Set an entry attribute."""
    if value and (not getattr(entry, attribute) or overwrite):
        entry = entry._replace(**{attribute: value})
    return entry


def merge_non_transaction_entries(imported_entries, enhanced_transactions):
    """Merge modified transactions back into a list of entries."""
    enhanced_entries = []
    enhanced_transactions_iter = iter(enhanced_transactions)
    for entry in imported_entries:
        # pylint: disable=isinstance-second-argument-not-valid-type
        if isinstance(entry, Transaction):
            enhanced_entries.append(next(enhanced_transactions_iter))
        else:
            enhanced_entries.append(entry)

    return enhanced_entries



"""Machine learning importer decorators."""

# pylint: disable=unsubscriptable-object

from __future__ import annotations

import logging
import threading
from typing import TYPE_CHECKING, Any, Callable

from beancount.core import data
from beancount.core.data import (
    Close,
    Open,
    Transaction,
    filter_txns,
)
from beancount.core.data import sorted as beancount_sorted
from beangulp.importer import Importer
from sklearn.pipeline import FeatureUnion, make_pipeline
from sklearn.svm import SVC

from smart_importer.entries import (
    merge_non_transaction_entries,
    set_entry_attribute,
)
from smart_importer.pipelines import get_pipeline

if TYPE_CHECKING:
    from sklearn import Pipeline

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ImporterWrapper(Importer):
    """Wrapper around an importer for enriching it with smart importer logic.

    Args:
        importer: The importer to wrap
        predictor: The entry predictor
    """

    def __init__(self, importer: Importer, predictor: EntryPredictor):
        self.importer = importer
        self.predictor = predictor

    @property
    def name(self) -> str:
        return self.importer.name

    def identify(self, filepath: str) -> bool:
        return self.importer.identify(filepath)

    def account(self, filepath):
        return self.importer.account(filepath)

    def date(self, filepath):
        return self.importer.date(filepath)

    def filename(self, filepath):
        return self.importer.filename(filepath)

    def deduplicate(self, entries, existing):
        return self.importer.deduplicate(entries, existing)

    def sort(self, entries, reverse=False):
        return self.importer.sort(entries, reverse)

    def extract(self, filepath, existing):
        entries = self.importer.extract(filepath, existing)
        account = self.importer.account(filepath)
        modified_entries = self.predictor.hook(
            [(filepath, entries, account, self.importer)], existing
        )
        return modified_entries[0][1]


class EntryPredictor:
    """Base class for machine learning importer helpers.

    Args:
        predict: Whether to add predictions to the entries.
        overwrite: When an attribute is predicted but already exists on an
            entry, overwrite the existing one.
        string_tokenizer: Tokenizer can let smart_importer support more
            languages. This parameter should be an callable function with
            string parameter and the returning should be a list.
        denylist_accounts: Transations with any of these accounts will be
            removed from the training data.
    """

    # pylint: disable=too-many-instance-attributes

    weights: dict[str, float] = {}
    attribute: str | None = None

    def __init__(
        self,
        predict: bool = True,
        overwrite: bool = False,
        string_tokenizer: Callable[[str], list] | None = None,
        denylist_accounts: list[str] | None = None,
    ) -> None:
        super().__init__()
        self.training_data = None
        self.open_accounts: dict[str, str] = {}
        self.denylist_accounts = set(denylist_accounts or [])
        self.pipeline: Pipeline | None = None
        self.is_fitted = False
        self.lock = threading.Lock()
        self.predict = predict
        self.overwrite = overwrite
        self.string_tokenizer = string_tokenizer

    def wrap(self, importer: Importer):
        """Wrap an existing importer with smart importer logic.

        Args:
            importer: The importer to wrap.
        """
        return ImporterWrapper(importer, self)

    def hook(
        self,
        imported_entries: list[
            tuple[str, data.Entries, data.Account, Importer]
        ],
        existing_entries: data.Directives,
    ) -> list[tuple[str, data.Entries, data.Account, Importer]]:
        """Predict attributes for imported transactions.

        Args:
            imported_entries: The list of imported entries.
            existing_entries: The list of existing entries as passed to the
                importer - will be used as training data.

        Returns:
            A list of entries, modified by this predictor.
        """
        with self.lock:
            all_entries = existing_entries or []
            self.load_open_accounts(all_entries)
            all_transactions = list(filter_txns(existing_entries))
            self.define_pipeline()

            result = []
            for filename, entries, account, importer in imported_entries:
                self.load_training_data(all_transactions, account)
                self.train_pipeline()
                result.append(
                    (
                        filename,
                        self.process_entries(entries),
                        account,
                        importer,
                    )
                )
            return result

    def load_open_accounts(self, existing_entries):
        """Return map of accounts which have been opened but not closed."""
        account_map = {}

        for entry in beancount_sorted(existing_entries):
            # pylint: disable=isinstance-second-argument-not-valid-type
            if isinstance(entry, Open):
                account_map[entry.account] = entry
            elif isinstance(entry, Close):
                account_map.pop(entry.account)

        self.open_accounts = account_map

    def load_training_data(self, all_transactions, account):
        """Load training data, i.e., a list of Beancount entries."""
        self.training_data = [
            txn
            for txn in all_transactions
            if self.training_data_filter(txn, account)
        ]
        if not self.training_data:
            if len(all_transactions) > 0:
                logger.warning(
                    "Cannot train the machine learning model"
                    "None of the training data matches the accounts"
                )
            else:
                logger.warning(
                    "Cannot train the machine learning model"
                    "No training data found"
                )
        else:
            logger.debug(
                "Loaded training data with %d transactions, "
                "filtered from %d total transactions",
                len(self.training_data),
                len(all_transactions),
            )

    def training_data_filter(self, txn, account):
        """Filter function for the training data."""
        found_import_account = False
        for pos in txn.postings:
            if pos.account not in self.open_accounts:
                return False
            if pos.account in self.denylist_accounts:
                return False
            if not account or pos.account.startswith(account):
                found_import_account = True

        return found_import_account

    @property
    def targets(self):
        """The training targets for the given training data.

        Returns:
            A list training targets (of the same length as the training data).
        """
        if not self.attribute:
            raise NotImplementedError
        return [
            getattr(entry, self.attribute) or ""
            for entry in self.training_data
        ]

    def define_pipeline(self) -> None:
        """Defines the machine learning pipeline based on given weights."""

        transformers = [
            (attribute, get_pipeline(attribute, self.string_tokenizer))
            for attribute in self.weights
        ]

        self.pipeline = make_pipeline(
            FeatureUnion(
                transformer_list=transformers, transformer_weights=self.weights
            ),
            SVC(kernel="linear"),
        )

    def train_pipeline(self) -> None:
        """Train the machine learning pipeline."""

        self.is_fitted = False
        targets_count = len(set(self.targets))

        if targets_count == 0:
            logger.warning(
                "Cannot train the machine learning model "
                "because there are no targets."
            )
        elif targets_count == 1:
            self.is_fitted = True
            logger.debug("Only one target possible.")
        else:
            assert self.pipeline is not None
            self.pipeline.fit(self.training_data, self.targets)
            self.is_fitted = True
            logger.debug("Trained the machine learning model.")

    def process_entries(
        self, imported_entries: data.Directives
    ) -> data.Directives:
        """Process imported entries.

        Transactions might be modified, all other entries are left as is.

        Returns:
            The list of entries to be imported.
        """
        enhanced_transactions = self.process_transactions(
            list(filter_txns(imported_entries))
        )
        return merge_non_transaction_entries(
            imported_entries, enhanced_transactions
        )

    def apply_prediction(
        self, entry: data.Transaction, prediction: Any
    ) -> data.Transaction:
        """Apply a single prediction to an entry.

        Args:
            entry: A Beancount entry.
            prediction: The prediction for an attribute.

        Returns:
            The entry with the prediction applied.
        """
        if not self.attribute:
            raise NotImplementedError
        return set_entry_attribute(
            entry, self.attribute, prediction, overwrite=self.overwrite
        )

    def process_transactions(
        self, transactions: list[Transaction]
    ) -> list[Transaction]:
        """Process a list of transactions."""
        if not self.is_fitted or not transactions:
            return transactions
        if self.predict:
            if len(set(self.targets)) == 1:
                transactions = [
                    self.apply_prediction(entry, self.targets[0])
                    for entry in transactions
                ]
                logger.debug("Apply predictions without pipeline")
            elif self.pipeline:
                predictions = self.pipeline.predict(transactions)
                transactions = [
                    self.apply_prediction(entry, prediction)
                    for entry, prediction in zip(transactions, predictions)
                ]
                logger.debug("Apply predictions with pipeline")
            logger.debug(
                "Added predictions to %d transactions",
                len(transactions),
            )

        return transactions



"""Tests for the Machine Learning Helpers."""

# pylint: disable=missing-docstring
import numpy as np
from beancount.parser import parser

from smart_importer.pipelines import (
    AttrGetter,
    NumericTxnAttribute,
    txn_attr_getter,
)

TEST_DATA, _, __ = parser.parse_string(
    """
2016-01-01 open Assets:US:BofA:Checking USD
2016-01-01 open Expenses:Food:Groceries USD
2016-01-01 open Expenses:Food:Coffee USD

2016-01-06 * "Farmer Fresh" "Buying groceries"
  Assets:US:BofA:Checking  -10.00 USD

2016-01-07 * "Starbucks" "Coffee"
  Assets:US:BofA:Checking  -4.00 USD
  Expenses:Food:Coffee

2016-01-07 * "Farmer Fresh" "Groceries"
  Assets:US:BofA:Checking  -11.20 USD
  Expenses:Food:Groceries

2016-01-08 * "Gimme Coffee" "Coffee"
  Assets:US:BofA:Checking  -3.50 USD
  Expenses:Food:Coffee
"""
)
TEST_TRANSACTIONS = TEST_DATA[3:]
TEST_TRANSACTION = TEST_TRANSACTIONS[0]


def test_get_payee():
    assert AttrGetter("payee").transform(TEST_TRANSACTIONS) == [
        "Farmer Fresh",
        "Starbucks",
        "Farmer Fresh",
        "Gimme Coffee",
    ]


def test_get_narration():
    assert AttrGetter("narration").transform(TEST_TRANSACTIONS) == [
        "Buying groceries",
        "Coffee",
        "Groceries",
        "Coffee",
    ]


def test_get_metadata():
    txn = TEST_TRANSACTION
    txn.meta["attr"] = "value"
    assert AttrGetter("meta.attr").transform([txn]) == ["value"]
    assert AttrGetter("meta.attr", "default").transform(TEST_TRANSACTIONS) == [
        "value",
        "default",
        "default",
        "default",
    ]


def test_get_day_of_month():
    get_day = txn_attr_getter("date.day")
    assert list(map(get_day, TEST_TRANSACTIONS)) == [6, 7, 7, 8]

    extract_day = NumericTxnAttribute("date.day")
    transformed = extract_day.transform(TEST_TRANSACTIONS)
    assert (transformed == np.array([[6], [7], [7], [8]])).all()



"""Smart importer for Beancount and Fava."""

from smart_importer.entries import update_postings
from smart_importer.predictor import EntryPredictor


class PredictPayees(EntryPredictor):
    """Predicts payees."""

    attribute = "payee"
    weights = {"narration": 0.8, "payee": 0.5, "date.day": 0.1}


class PredictPostings(EntryPredictor):
    """Predicts posting accounts."""

    weights = {"narration": 0.8, "payee": 0.5, "date.day": 0.1}

    @property
    def targets(self):
        return [
            " ".join(sorted(posting.account for posting in txn.postings))
            for txn in self.training_data
        ]

    def apply_prediction(self, entry, prediction):
        return update_postings(entry, prediction.split(" "))



"""Machine learning pipelines for data extraction."""

from __future__ import annotations

import operator

import numpy
from beancount.core.data import Transaction
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.pipeline import make_pipeline


class NoFitMixin:
    """Mixin that implements a transformer's fit method that returns self."""

    def fit(self, *_, **__):
        """A noop."""
        return self


def txn_attr_getter(attribute_name: str):
    """Return attribute getter for a transaction that also handles metadata."""
    if attribute_name.startswith("meta."):
        meta_attr = attribute_name[5:]

        def getter(txn):
            return txn.meta.get(meta_attr)

        return getter
    return operator.attrgetter(attribute_name)


class NumericTxnAttribute(BaseEstimator, TransformerMixin, NoFitMixin):
    """Get a numeric transaction attribute and vectorize."""

    def __init__(self, attr: str):
        self.attr = attr
        self._txn_getter = txn_attr_getter(attr)

    def transform(self, data: list[Transaction], _y=None):
        """Return list of entry attributes."""
        return numpy.array([self._txn_getter(d) for d in data], ndmin=2).T


class AttrGetter(BaseEstimator, TransformerMixin, NoFitMixin):
    """Get a string transaction attribute."""

    def __init__(self, attr: str, default: str | None = None):
        self.attr = attr
        self.default = default
        self._txn_getter = txn_attr_getter(attr)

    def transform(self, data: list[Transaction], _y=None):
        """Return list of entry attributes."""
        return [self._txn_getter(d) or self.default for d in data]


class StringVectorizer(CountVectorizer):
    """Subclass of CountVectorizer that handles empty data."""

    def __init__(self, tokenizer=None):
        super().__init__(ngram_range=(1, 3), tokenizer=tokenizer)

    def fit_transform(self, raw_documents: list[str], y=None):
        try:
            return super().fit_transform(raw_documents, y)
        except ValueError:
            return numpy.zeros(shape=(len(raw_documents), 0))

    def transform(self, raw_documents: list[str], _y=None):
        try:
            return super().transform(raw_documents)
        except ValueError:
            return numpy.zeros(shape=(len(raw_documents), 0))


def get_pipeline(attribute: str, tokenizer):
    """Make a pipeline for a given entry attribute."""

    if attribute.startswith("date."):
        return NumericTxnAttribute(attribute)

    # Treat all other attributes as strings.
    return make_pipeline(
        AttrGetter(attribute, default=""), StringVectorizer(tokenizer)
    )



"""Tests for the entry helpers."""

from __future__ import annotations

# pylint: disable=missing-docstring
from beancount.parser import parser

from smart_importer.entries import update_postings

TEST_DATA, _errors, _options = parser.parse_string(
    """
2016-01-06 * "Farmer Fresh" "Buying groceries"
  Assets:US:BofA:Checking  -10.00 USD

2016-01-06 * "Farmer Fresh" "Buying groceries"
  Assets:US:BofA:Checking  -10.00 USD
  Assets:US:BofA:Checking   10.00 USD
"""
)


def test_update_postings() -> None:
    txn0 = TEST_DATA[0]

    def _update(accounts: list[str]) -> list[tuple[str, bool]]:
        """Update, get accounts and whether this is the original posting."""
        updated = update_postings(txn0, accounts)
        return [(p.account, p is txn0.postings[0]) for p in updated.postings]

    assert _update(["Assets:US:BofA:Checking", "Assets:Other"]) == [
        ("Assets:US:BofA:Checking", True),
        ("Assets:Other", False),
    ]

    assert _update(
        ["Assets:US:BofA:Checking", "Assets:US:BofA:Checking", "Assets:Other"]
    ) == [
        ("Assets:US:BofA:Checking", True),
        ("Assets:US:BofA:Checking", False),
        ("Assets:Other", False),
    ]

    assert _update(["Assets:Other", "Assets:Other2"]) == [
        ("Assets:Other", False),
        ("Assets:Other2", False),
        ("Assets:US:BofA:Checking", True),
    ]

    txn1 = TEST_DATA[1]
    assert update_postings(txn1, ["Assets:Other"]) == txn1
